import imp
import os
import types
import sys
from pyrthon._transformer import transform


class PyrsistentImporter(object):
    def __init__(self):
        self._match_expressions = []

    def module_matches(self, name):
        for expr in self._match_expressions:
            if callable(expr):
                if expr(name):
                    return True

            else:
                if expr.endswith('*') and name.startswith(expr[:-1]):
                    return True

                if expr == name:
                    return True

        return False

    def add_matchers(self, matchers):
        self._match_expressions.extend(matchers)

    def find_module(self, name, path=None):
        if self.module_matches(name):
            return self

        return None

    def load_module(self, name):
        try:
            suffix = name.split('.')[-1]
            fd, pathname, (suffix, mode, type_) = imp.find_module(suffix)
        except ImportError:
            return None

        module = types.ModuleType(name) #create empty module object

        with fd:
            if type_ == imp.PY_SOURCE:
                filename = pathname
            elif type_ == imp.PY_COMPILED:
                filename = pathname[:-1]
            elif type_ == imp.PKG_DIRECTORY:
                filename = os.path.join(pathname, '__init__.py')
                module.__path__ = [pathname]
            else:
                return imp.load_module(name, fd, pathname, (suffix, mode, type_))

            if filename != pathname:
                try:
                    with open(filename, 'U') as real_file:
                        src = real_file.read()
                except IOError: #fallback
                    return imp.load_module(name, fd, pathname, (suffix, mode, type_))
            else:
                src = fd.read()

        module.__file__ = filename
        inlined = transform(src)
        code = compile(inlined, filename, 'exec')
        sys.modules[name] = module
        exec(code,  module.__dict__)

        return module

importer = PyrsistentImporter()





import ast


class PyrsistentTransformer(ast.NodeTransformer):
    def visit_Module(self, node):
        node = self.generic_visit(node)
        import_node = ast.ImportFrom(module='pyrsistent', names=[ast.alias(name='pvector'), ast.alias(name='pmap'), ast.alias(name='pset')], level=0)
        node.body.insert(0, import_node)
        return node

    def _visit_collection(self, node, pname):
        node = self.generic_visit(node)
        name_node = ast.Name(id=pname, ctx=ast.Load())
        new_node = ast.Call(func=name_node, args=[node], keywords=[])
        return new_node

    @staticmethod
    def _is_empty_set_call(node):
        return isinstance(node.func, ast.Name) and node.func.id == 'set' and not node.args

    def visit_Call(self, node):
        if self._is_empty_set_call(node):
            name_node = ast.Name(id='pset', ctx=ast.Load())
            return ast.Call(func=name_node, args=[], keywords=[])

        return self.generic_visit(node)

    def visit_Dict(self, node):
        return self._visit_collection(node, 'pmap')

    def visit_DictComp(self, node):
        return self._visit_collection(node, 'pmap')

    def visit_List(self, node):
        return self._visit_collection(node, 'pvector')

    def visit_ListComp(self, node):
        return self._visit_collection(node, 'pvector')

    def visit_Set(self, node):
        return self._visit_collection(node, 'pset')

    def visit_SetComp(self, node):
        return self._visit_collection(node, 'pset')


def transform(src):
    """ Transforms the given source to use pvectors, pmaps and psets to replace built in structures """
    tree = ast.parse(src)
    transformer = PyrsistentTransformer()
    new_tree = transformer.visit(tree)
    ast.fix_missing_locations(new_tree)
    return new_tree


import sys
import pyrthon._import_hook as import_hook
from pyrthon.pyrthon_extension import load_ipython_extension,  unload_ipython_extension


def register(*matchers):
    """
    Register modules for literal substitution. Matchers can be specified in three ways:

    * Exact module name.
      register('foo.bar')
    * Prefix with wild card. All submodules of the prefix.
      register('foo.*')
    * Custom matcher function that returns true if substitution should be applied in the module.
      register(lambda name: name.startswith('foo') and name.endswith('baz'))

    This function must be called before the referred modules are imported.
    """
    if import_hook.importer not in sys.meta_path:
        sys.meta_path.insert(0, import_hook.importer)

    import_hook.importer.add_matchers(matchers)



import code
import ast
from pyrthon._transformer import transform


class PyrsistentConsole(code.InteractiveConsole):
    def runsource(self, source, filename='<input>', symbol='single'):
        try:
            code_object = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            code_object = ''

        if code_object is None:
            # This means it's incomplete
            return True

        try:
            tree = transform(source)
            tree = ast.Interactive(tree.body)
            code_object = compile(tree, filename, symbol, self.compile.compiler.flags, 1)
        except (OverflowError, SyntaxError, ValueError):
            # This means there's a syntax error
            self.showsyntaxerror(filename)
            return False

        # This means it was successfully compiled; `runcode` takes care of any runtime failures
        self.runcode(code_object)
        return False



from pyrthon._transformer import PyrsistentTransformer

def load_ipython_extension(ipython):
    ''' load extension for prython in ipython'''
    pyr_transform = PyrsistentTransformer()
    ipython.ast_transformers.append(pyr_transform)
    print("'pyrthon' extension loaded ")

def unload_ipython_extension(ipython):
    ''' unload the pyrthon ipython extension'''
    ipython.ast_transformers = [transformer 
            for transformer in ipython.ast_transformers 
            if not isinstance(transformer, PyrsistentTransformer)]
    print("'pyrthon' extension unloaded")



import os
from setuptools import setup
from pyrthon._version import __version__


f = open(os.path.join(os.path.dirname(__file__), 'README.rst'))
readme = f.read()
f.close()


setup(
    name='pyrthon',
    version=__version__,
    description='Literal syntax for Pyrsistent data structures',
    long_description=readme,
    author='Tobias Gustafsson',
    author_email='tobias.l.gustafsson@gmail.com',
    url='http://github.com/tobgu/pyrthon/',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    test_suite='tests',
    scripts=[],
    install_requires=['pyrsistent'],
    packages=['pyrthon']
)
